//------------------------------------------------------------------------------
// Copyright (c) 2004, Perforce Software, Inc.  All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1.  Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
// 
// 2.  Redistributions in binary form must reproduce the above copyright
//     notice, this list of conditions and the following disclaimer in the
//     documentation and/or other materials provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL PERFORCE SOFTWARE, INC. BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//------------------------------------------------------------------------------
#define NEED_STAT
#include <clientapi.h>
#include <strtable.h>
#include <vararray.h>
#include "filelist.h"
#include "adminuser.h"

AdminUser::AdminUser()
{ 
    md 		= new StrBufDict;
    files	= new FileList;
}

AdminUser::~AdminUser()
{
    delete md;
    delete files;
}

//
// OutputStat is only called when we run our "p4 fstat", not when we run
// the command itself.
//
void
AdminUser::OutputStat( StrDict *varList )
{
    StrPtr *path 	= varList->GetVar( "clientFile" );
    StrPtr *mask 	= varList->GetVar( "attr-unixMask" );
    StrPtr *act		= varList->GetVar( "headAction" );

    if( !path || !mask ) return;
    if( act && *act == "delete" ) return;
    if( GetMode( varList, mode ) )
	SetMode( path, &mode );
}


void
AdminUser::SetMode( StrPtr *path, StrPtr *mode )
{
    md->SetVar( path->Text(), mode->Text() );
}

void
AdminUser::PushFile( StrPtr *file )
{
    files->Put( file );
}

void
AdminUser::PopFile()
{
    files->DeleteLast();
}


int
AdminUser::GetMode( StrDict *varList, StrBuf &mode )
{
    StrPtr	*mask	= varList->GetVar( "attr-unixMask" );
    if( !mask ) return 0;

    // Start by making the file readable.
    int	m	= BM_RD;

    // If the file should be left writable, or in the writable bits.
    if( IsWritable( varList ) )
	m |= BM_WR;

    // If the file should be executable, or in the exec bits.
    if( IsExecutable( varList ) )
	m |= BM_EX;

    // Now mask out the bits the user doesn't want.
    m = m & ~ strtol( mask->Text(), 0, 8 );

    // If the mode string is less than 5 chars long, then we'll also mask
    // out the high bits of the permissions to prevent accidents. Those that
    // know what they're doing and want setuid/setgid or the sticky bit set
    // can use masks like 03022 etc.
    if( mask->Length() < 5 )
	m = m & 0777;

    // Update the mode string.
    mode.Alloc( 8 );
    snprintf( mode.Text(), 8, "%05o", m );
    mode.SetLength();
    return 1;
}

int
AdminUser::IsExecutable( StrDict *varList )
{
    StrPtr	*type	= GetFileType( varList );

    if( !type ) 
	return 0;

    if( type->Contains( StrRef( "xtext" ) ) || 
	type->Contains( StrRef( "xbinary" ) ) )
	return 1;

    const char *pos = type->Contains( StrRef( "+" ) );
    if( !pos ) return 0;

    pos++;
    if( strchr( pos, 'x' ) )
	return 1;
    return 0;
}


int
AdminUser::IsWritable( StrDict *varList )
{
    StrPtr	*type	= GetFileType( varList );

    if( !type ) 
	return 0;

    const char *pos = type->Contains( StrRef( "+" ) );
    if( !pos ) return 0;

    pos++;
    if( strchr( pos, 'w' ) )
	return 1;
    return 0;
}

StrPtr *
AdminUser::GetFileType( StrDict *varList )
{
    return varList->GetVar( "headType" );
}



